import os
import time

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.api import BaseObserver

from argparse import ArgumentParser, Namespace

from ontol import Parser, JSONSerializer, PlantUML


VERSION: str = os.getenv('ONTOL_VERSION', 'dev')


class CLI:
    def __init__(self) -> None:
        self.args_parser: ArgumentParser = ArgumentParser(
            description='Ontol DSL Parser - A tool for parsing and visualizing ontology files written in the Ontol DSL.'
        )
        self.args_parser.add_argument(
            'file', type=str, help='Path to the .ontol file to be parsed.'
        )
        self.args_parser.add_argument(
            '-w',
            '--watch',
            action='store_true',
            help='Watch the specified file for changes and re-parse automatically.',
        )
        self.args_parser.add_argument(
            '-v',
            '--version',
            action='version',
            version=f'%(prog)s {VERSION}',
            help='Show the version of the program and exit.',
        )

        self.parser: Parser = Parser()
        self.serializer: JSONSerializer = JSONSerializer()
        self.plantuml: PlantUML = PlantUML()

    def run(self) -> None:
        args: Namespace = self.args_parser.parse_args()

        if args.watch:
            self.watch_file(args.file)
        else:
            self.parse_file(args.file)

    def parse_file(self, file_path: str) -> None:
        try:
            with open(file_path, 'r', encoding='utf-8') as file:
                content: str = file.read()
                ontology, warnings = self.parser.parse(content, file_path)

                # Print warnings
                if warnings:
                    print('\n\n'.join(warnings))

                # JSON
                json_content: str = self.serializer.serialize(ontology)
                json_file_path: str = os.path.splitext(file_path)[0] + '.json'
                with open(json_file_path, 'w', encoding='utf-8') as json_file:
                    json_file.write(json_content)

                # PlantUML
                plantuml_content: str = self.plantuml.generate(ontology)
                puml_file_path: str = os.path.splitext(file_path)[0] + '.puml'
                with open(puml_file_path, 'w', encoding='utf-8') as puml_file:
                    puml_file.write(plantuml_content)

                self.plantuml.processes_puml_to_png(puml_file_path)
        except Exception as e:
            print(e)

    def watch_file(self, file_path):
        self.parse_file(file_path)

        class FileChangeHandler(FileSystemEventHandler):
            def __init__(self, parse_callback):
                super().__init__()
                self.parse_callback = parse_callback

            def on_modified(self, event):
                if event.src_path.endswith('.ontol'):
                    print(f'File {event.src_path} modified, re-parsing...')
                    self.parse_callback(event.src_path)

        event_handler: FileChangeHandler = FileChangeHandler(self.parse_file)

        observer: BaseObserver = Observer()
        observer.schedule(event_handler, path=file_path, recursive=True)
        observer.start()

        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()


def main():
    cli: CLI = CLI()
    cli.run()


if __name__ == '__main__':
    main()



from src.ontol import Function, Meta, Ontology, Relationship, Term


def test_term_creation() -> None:
    term: Term = Term(name='test_term', label='TestTerm', description='A test term')
    assert term.name == 'test_term'
    assert term.label == 'TestTerm'
    assert term.description == 'A test term'
    assert (
        repr(term)
        == 'Term(name=test_term, label=TestTerm, description=A test term, attributes={})'
    )


def test_term_with_attributes_creation() -> None:
    attributes = {'color': 'red'}
    term_with_attributes: Term = Term(
        name='test_term',
        label='TestTerm',
        description='A test term',
        attributes=attributes,
    )
    assert term_with_attributes.attributes == attributes
    assert (
        repr(term_with_attributes)
        == f'Term(name=test_term, label=TestTerm, description=A test term, attributes={attributes})'
    )


def test_term_with_empty_label_creation() -> None:
    term: Term = Term(name='test_term', label='', description='A test term')
    assert term.label == ''
    assert (
        repr(term)
        == 'Term(name=test_term, label=, description=A test term, attributes={})'
    )


def test_term_with_empty_description_creation() -> None:
    term: Term = Term(name='test_term', label='TestTerm', description='')
    assert term.description == ''
    assert (
        repr(term)
        == 'Term(name=test_term, label=TestTerm, description=, attributes={})'
    )


def test_function_creation():
    func: Function = Function(
        name='sum',
        label='Sum',
        input_types=[('int', 'First number'), ('int', 'Second number')],
        output_type=('int', 'Result'),
    )
    assert func.name == 'sum'
    assert func.label == 'Sum'
    assert func.input_types == [('int', 'First number'), ('int', 'Second number')]
    assert func.output_type == ('int', 'Result')
    assert (
        repr(func)
        == "Function(name=sum, label=Sum, input_types=[('int', 'First number'), ('int', 'Second number')], output_type=('int', 'Result'), attributes={})"
    )


def test_function_wth_empty_label_creation():
    func: Function = Function(
        name='sum',
        label='',
        input_types=[('int', 'First number'), ('int', 'Second number')],
        output_type=('int', 'Result'),
    )
    assert func.label == ''
    assert (
        repr(func)
        == "Function(name=sum, label=, input_types=[('int', 'First number'), ('int', 'Second number')], output_type=('int', 'Result'), attributes={})"
    )


def test_function_wth_empty_input_types_creation():
    func: Function = Function(
        name='sum',
        label='Sum',
        input_types=[],
        output_type=('int', 'Result'),
    )
    assert func.input_types == []
    assert (
        repr(func)
        == "Function(name=sum, label=Sum, input_types=[], output_type=('int', 'Result'), attributes={})"
    )


def test_relationship_creation() -> None:
    rel: Relationship = Relationship(
        parent='set', relationship='composite', child=['element']
    )
    assert rel.parent == 'set'
    assert rel.relationship == 'composite'
    assert rel.child == ['element']
    assert (
        repr(rel)
        == "Relationship(parent=set, relationship=composite, child=['element'], attributes={})"
    )


def test_meta_creation():
    meta: Meta = Meta(
        version='1.0',
        name='TestOntology',
        author='Author',
        description='A test ontology',
        type='Base',
        date_created='2024-01-01',
    )
    assert meta.version == '1.0'
    assert meta.name == 'TestOntology'
    assert meta.author == 'Author'
    assert meta.description == 'A test ontology'
    assert meta.type == 'Base'
    assert meta.date_created == '2024-01-01'
    assert (
        repr(meta)
        == 'Meta(version=1.0, name=TestOntology, author=Author, description=A test ontology, type=Base, date_created=2024-01-01)'
    )


def test_meta_with_empty_description_creation():
    meta_no_optional: Meta = Meta(
        version='1.0',
        name='TestOntology',
        author='Author',
        description=None,
        type='Base',
        date_created='2024-01-01',
    )
    assert meta_no_optional.description is None
    assert (
        repr(meta_no_optional)
        == 'Meta(version=1.0, name=TestOntology, author=Author, description=None, type=Base, date_created=2024-01-01)'
    )


def test_ontology_operations():
    ontology: Ontology = Ontology()
    char: Term = Term(name='str', label='String', description='')
    string: Term = Term(name='string', label='String', description='')
    concatanate: Function = Function(
        name='concatanate',
        label='Concatanate',
        input_types=[('str', ''), ('str', '')],
        output_type=('str', 'Result'),
    )
    rel: Relationship = Relationship(
        parent='string', relationship='composition', child='char'
    )
    meta: Meta = Meta(
        version='1.0',
        name='StringOntology',
        author='Author',
        description='A string ontology',
        date_created='2024-01-01',
    )

    ontology.add_type(char)
    ontology.add_type(string)
    ontology.add_function(concatanate)
    ontology.add_relationship(rel)
    ontology.set_meta(meta)

    assert ontology.types == [char, string]
    assert ontology.functions == [concatanate]
    assert ontology.hierarchy == [rel]
    assert ontology.meta == meta

    assert repr([char, string]) in repr(ontology)
    assert repr([concatanate]) in repr(ontology)
    assert repr([rel]) in repr(ontology)
    assert repr(meta) in repr(ontology)



import json

from dataclasses import asdict

from ontol import Ontology, Term, Function, Meta, Relationship


class JSONSerializer:
    @staticmethod
    def serialize(ontology: Ontology) -> str:
        data: dict[str, object] = {
            'terms': [JSONSerializer._serialize_term(t) for t in ontology.types],
            'functions': [
                JSONSerializer._serialize_function(f) for f in ontology.functions
            ],
            'hierarchy': [
                JSONSerializer._serialize_relationship(r) for r in ontology.hierarchy
            ],
            'meta': JSONSerializer._serialize_meta(ontology.meta)
            if ontology.meta
            else None,
        }
        return json.dumps(data, ensure_ascii=False, indent=4)

    @staticmethod
    def _serialize_term(type_def: Term) -> dict[str, str | None]:
        return asdict(type_def)

    @staticmethod
    def _serialize_function(func_def: Function) -> dict[str, str | list[str] | None]:
        return asdict(func_def)

    @staticmethod
    def _serialize_meta(meta: Meta) -> dict[str, str | None]:
        return asdict(meta)

    @staticmethod
    def _serialize_relationship(
        rel_def: Relationship,
    ) -> dict[str, str | list[str] | None]:
        return asdict(rel_def)



from src.ontol import Function, Meta, Ontology, Relationship, Term, PlantUML

import pytest


@pytest.fixture
def mock_ontology():
    ontology = Ontology()
    ontology.set_meta(
        Meta(
            version='1.0',
            name='TestOntology',
            author='Author',
            description='Test ontology',
            type='Базовый',
            date_created='2024-01-01',
        )
    )
    # TODO: add term without color attribute
    ontology.add_type(Term(name='MyTypeParent', label='test label Term1', description='A test type1',
                           attributes={'color': '#E6B8B7'}))
    ontology.add_type(Term(name='MyTypeChild', label='test label Term2', description='A test type2'))
    ontology.add_function(
        Function(
            name='MyFunction1',
            label='test label Func1',
            input_types=[('int', 'test1 in1'), ('str', 'test1 in2')],
            output_type=('bool', 'test_type1'),
            attributes={'colorArrow': '#E6B8B7'}
        )
    )
    ontology.add_function(
        Function(
            name='MyFunction2',
            label='test label Func2',
            input_types=[('str', 'test2 in1'), ('float', 'test2 in2')],
            output_type=('str', 'test_type2')
        )
    )
    ontology.add_relationship(
        Relationship(
            parent='MyTypeParent',
            relationship='composition',
            child=['MyTypeChild'],
        )
    )
    return ontology


@pytest.fixture
def generator():
    return PlantUML()


def test_generate_full_uml(generator: PlantUML, mock_ontology):
    uml_output = generator.generate(mock_ontology)
    assert '@startuml' in uml_output
    assert '@enduml' in uml_output

    assert 'MyTypeParent' in uml_output
    assert 'test label Term1' in uml_output
    assert '(A test type1)' in uml_output
    assert '#E6B8B7' in uml_output
    assert 'MyTypeChild' in uml_output
    assert 'test label Term2' in uml_output
    assert '(A test type2)' in uml_output
    assert '#white' in uml_output

    assert ('"test label Func1\\n(int: test1 in1, str: test1 in2 -> bool: test_type1)" as MyFunction1 #white'
            in uml_output)
    assert ('"test label Func2\\n(str: test2 in1, float: test2 in2 -> str: test_type2)" as MyFunction2 #white'
            in uml_output)
    print(uml_output)
    assert 'int  --[#E6B8B7]->  MyFunction1 ' in uml_output
    assert 'str  --[#E6B8B7]->  MyFunction1 ' in uml_output
    assert 'MyFunction1  --[#E6B8B7]->  bool ' in uml_output

    assert 'str  --[#black]->  MyFunction2 ' in uml_output
    assert 'float  --[#black]->  MyFunction2 ' in uml_output
    assert 'MyFunction2  --[#black]->  str ' in uml_output

    assert 'MyTypeParent  --[#black]-*  MyTypeChild ' in uml_output


def test_generate_type(generator: PlantUML):
    term = Term(name='MyType', description='A test type', label='test label')
    result = generator._generate_type(term)
    assert result == 'class MyType {\n  A test type\n}'


def test_generate_function(generator: PlantUML):
    func = Function(
        name='MyFunction',
        label='test label',
        input_types=[('int', 'test')],
        output_type=('bool', "test"),
        attributes={'color': '#E6B8B7'}
    )
    result = generator._generate_function(func)
    assert 'MyFunction' in result
    assert 'class MyFunction <<Function>> {' in result
    assert '+MyFunction(int) : (bool)' in result
    assert 'test label' in result
    assert '}' in result


def test_generate_logical_expression(generator: PlantUML):
    relation = Relationship(
        parent='MyTypeParent',
        relationship='containsAAAA',
        child=['MyTypeChild'],
    )
    result = generator._generate_relationship(relation)
    print(result)
    assert 'MyTypeParent' in result
    assert 'containsAAAA' in result
    assert 'MyTypeChild' in result
    assert 'as' in result


import collections
import os
import zlib

import requests

from ontol import Function, Ontology, Relationship, Term


# TODO: make look like in technical task
class PlantUML:
    SERVER_URL = 'http://www.plantuml.com/plantuml/img/'

    def __init__(self, url=SERVER_URL):
        self.url = url

    def generate(self, ontology: Ontology) -> str:
        if ontology.meta.type == 'Базовый':
            return self._generate_novikov(ontology)
        else:
            return self._generate_base(ontology)

    def _generate_base(self, ontology: Ontology) -> str:
        uml_lines: list[str] = ['@startuml', 'skinparam classAttributeIconSize 0']

        if ontology.meta:
            uml_lines.append(f'title {ontology.meta.name} by {ontology.meta.author}')

        for term in ontology.types:
            uml_lines.append(self._generate_type(term))

        for function in ontology.functions:
            uml_lines.append(self._generate_function(function))

        for relationship in ontology.hierarchy:
            uml_lines.append(self._generate_relationship(relationship))

        uml_lines.append('@enduml')
        return '\n'.join(uml_lines)

    def _generate_novikov(self, ontology: Ontology) -> str:
        uml_lines: list[str] = [
            '@startuml',
            'skinparam backgroundColor #F0F8FF',
            'skinparam defaultTextAlignment center',
            'skinparam shadowing false',
            'skinparam dpi 150',
            'skinparam linetype ortho',
            'skinparam ranksep 40',
            'skinparam nodesep 30',
        ]

        if not ontology.meta:
            raise ValueError('No meta defined for ontology')

        uml_lines.append(f'package "{ontology.meta.name}" {{')

        for term in ontology.types:
            uml_lines.append(self._generate_rectangle(term))

        for function in ontology.functions:
            uml_lines.append(
                self._generate_rectangle(self.__prepare_function_term(function))
            )

        for function in ontology.functions:
            for relations in self.__prepare_function_hierarchy(function):
                uml_lines.append(self._generate_base_hierarchy(relations))

        for relationship in ontology.hierarchy:
            uml_lines.append(self._generate_base_hierarchy(relationship))

        uml_lines.append('}')
        uml_lines.append('@enduml')
        return '\n'.join(uml_lines)

    def _generate_rectangle(self, term: Term) -> str:
        return (
            f'rectangle "{term.label}'
            + (f'\\n({term.description})' if term.description else '')
            + f'" as {term.name} {term.attributes.get("color", "#white")}'
        )

    def _generate_base_hierarchy(self, relationship: Relationship) -> str:
        relationships = {
            'depends': {
                'forward': '...>',
                'backward': '<...',
                'bidirectional': '<...>',
            },
            'association': {
                'forward': '---',
                'backward': '---',
                'bidirectional': '---',
            },
            'directAssociation': {
                'forward': '--->',
                'backward': '<---',
                'bidirectional': '<--->',
            },
            'inheritance': {
                'forward': '---|>',
                'backward': '<|---',
                'bidirectional': '<|---|>',
            },
            'realization': {
                'forward': '...|>',
                'backward': '<|...',
                'bidirectional': '<|...|>',
            },
            'aggregation': {
                'forward': '---o',
                'backward': 'o---',
                'bidirectional': 'o---o',
            },
            'composition': {
                'forward': '---*',
                'backward': '*---',
                'bidirectional': '*---*',
            },
            'usage': {'forward': '...>', 'backward': '<...', 'bidirectional': '<...>'},
        }
        leftchar = (
            '"' + relationship.attributes['leftChar'] + '"'
        ) if relationship.attributes.get('leftChar') else ''
        rightchar = (
             '"' + relationship.attributes['rightChar'] + '"'
        ) if relationship.attributes.get('rightChar') else ''
        title = (
            ': "' + relationship.attributes['title'] + '"'
        ) if relationship.attributes.get('title') else ''
        color = '[' + relationship.attributes.get('color', '#black') + ']'
        relation = (
            relationships[relationship.relationship][
                relationship.attributes.get('direction', 'forward')
            ][:2]
            + color
            + relationships[relationship.relationship][
                relationship.attributes.get('direction', 'forward')
            ][2:]
        )
        res = ''
        res += (
            f'{relationship.parent} {leftchar} '
            f'{relation} '
            f'{rightchar} '
            f'{relationship.child[0]} {title}'
        ) + '\n'
        return res

    def __prepare_function_term(self, function: Function):
        input = [
            f'{el1}: {el2}' if el2 else f'{el1}' for el1, el2 in function.input_types
        ]
        output = (
            (f'{function.output_type[0]}: {function.output_type[1]}')
            if function.output_type[1]
            else f'{function.output_type[0]}'
        )
        desc = f'{", ".join(input)} -> {output}'
        return Term(
            function.name,
            function.label,
            desc,
            {'color': function.attributes.get('color', '#white')},
        )

    def __prepare_function_hierarchy(self, function: Function):
        relations = []

        input = collections.defaultdict(int)
        for type, _ in function.input_types:
            input[type] += 1

        for k, v in input.items():
            attributes_dict = {
                'color': function.attributes.get('colorArrow', '#black'),
                'title': function.attributes.get('inputTitle', ''),
                'leftChar': f'{v if v != 1 else ""}',
                'rightChar': '',
                'direction': 'forward',
            }
            relations.append(
                Relationship(
                    k,
                    function.attributes.get('type', 'directAssociation'),
                    [function.name],
                    attributes_dict,
                )
            )
        attributes_dict = {
            'color': function.attributes.get('colorArrow', '#black'),
            'title': function.attributes.get('outputTitle', ''),
            'leftChar': '',
            'rightChar': '',
            'direction': 'forward',
        }
        relations.append(
            Relationship(
                function.name,
                function.attributes.get('type', 'directAssociation'),
                [function.output_type[0]],
                attributes_dict,
            )
        )
        return relations

    def _generate_type(self, term: Term) -> str:
        return f'class {term.name} {{\n  {term.description}\n}}'

    # TODO: add comments to input and output types and add color for block
    def _generate_function(self, function: Function) -> str:
        inputs: str = ', '.join(map(lambda t: t[0], function.input_types))
        outputs: str = function.output_type[0]
        return (
            f'class {function.name} <<Function>> {{\n'
            f'  +{function.name}({inputs}) : ({outputs})\n'
            f'  {function.label}\n}}'
        )

    # TODO: check type of relationship
    def _generate_relationship(self, relationship: Relationship) -> str:
        return f'note "{relationship.parent} {relationship.relationship} {relationship.child}" as N{hash(relationship.parent) % 10000}'

    def processes_puml_to_png(self, puml_file):
        outfile = os.path.splitext(puml_file)[0] + '.png'

        with open(puml_file, 'r', encoding='utf-8') as file:
            plantuml_text = file.read()

        data = zlib.compress(plantuml_text.encode('utf-8'))[2:-4]
        encoded_text = ''
        for i in range(0, len(data), 3):
            if i + 2 == len(data):
                encoded_text += self.__encode3bytes(data[i], data[i + 1], 0)
            elif i + 1 == len(data):
                encoded_text += self.__encode3bytes(data[i], 0, 0)
            else:
                encoded_text += self.__encode3bytes(data[i], data[i + 1], data[i + 2])

        url = f'{self.url}{encoded_text}'
        response = requests.get(url)

        if response.status_code == 200:
            with open(outfile, 'wb') as out:
                out.write(response.content)

    @staticmethod
    def __encode3bytes(b1, b2, b3):
        chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_'
        c1 = b1 >> 2
        c2 = ((b1 & 0x3) << 4) | (b2 >> 4)
        c3 = ((b2 & 0xF) << 2) | (b3 >> 6)
        c4 = b3 & 0x3F
        return chars[c1] + chars[c2] + chars[c3] + chars[c4]



import json
from src.ontol import Function, Meta, Ontology, Relationship, Term, JSONSerializer

import pytest


@pytest.fixture
def sample_ontology():
    ontology = Ontology()
    ontology.add_type(
        Term(name='MyType', description='A sample type', label='A some label')
    )
    ontology.add_function(
        Function(
            name='MyFunction',
            label='Function',
            input_types=['int'],
            output_type=['bool'],
        )
    )
    ontology.add_relationship(
        Relationship(
            relationship='',
            parent='',
            child='',
        )
    )
    ontology.set_meta(
        Meta(
            version='1.0',
            name='SampleOntology',
            author='Author',
            description='Sample description',
            type='Базовый',
            date_created='2024-01-01',
        )
    )
    return ontology


@pytest.fixture
def serializer():
    return JSONSerializer()


def test_serialize_ontology(serializer, sample_ontology):
    json_output = serializer.serialize(sample_ontology)
    data = json.loads(json_output)

    assert 'terms' in data
    assert 'functions' in data
    assert 'hierarchy' in data
    assert 'meta' in data
    assert data['terms'] == [
        {
            'name': 'MyType',
            'label': 'A some label',
            'description': 'A sample type',
            'attributes': {},
        }
    ]
    assert data['functions'] == [
        {
            'name': 'MyFunction',
            'label': 'Function',
            'input_types': ['int'],
            'output_type': ['bool'],
            'attributes': {},
        }
    ]

    assert data['hierarchy'] == [
        {'parent': '', 'relationship': '', 'child': '', 'attributes': {}}
    ]
    assert data['meta'] == {
        'version': '1.0',
        'name': 'SampleOntology',
        'author': 'Author',
        'description': 'Sample description',
        'type': 'Базовый',
        'date_created': '2024-01-01',
    }


def test_serialize_empty_ontology(serializer):
    ontology = Ontology()
    json_output = serializer.serialize(ontology)
    data = json.loads(json_output)
    assert data == {
        'terms': [],
        'functions': [],
        'hierarchy': [],
        'meta': {
            'version': None,
            'name': None,
            'author': None,
            'description': None,
            'type': None,
            'date_created': None,
        },
    }


def test_serialize_term(serializer):
    term = Term(name='TestTerm', description='A test term', label='A test label')
    serialized_term = serializer._serialize_term(term)
    assert serialized_term == {
        'name': 'TestTerm',
        'label': 'A test label',
        'description': 'A test term',
        'attributes': {},
    }


def test_serialize_function(serializer):
    func = Function(
        name='TestFunction',
        input_types=['int'],
        output_type=['str'],
        label='A test function',
    )
    serialized_function = serializer._serialize_function(func)
    print(serialized_function)
    assert serialized_function == {
        'name': 'TestFunction',
        'label': 'A test function',
        'input_types': ['int'],
        'output_type': ['str'],
        'attributes': {},
    }


def test_serialize_meta(serializer):
    meta = Meta(
        version='1.0',
        name='TestOntology',
        author='Author',
        description='A test ontology',
        type='Базовый',
        date_created='2024-01-01',
    )
    serialized_meta = serializer._serialize_meta(meta)
    assert serialized_meta == {
        'version': '1.0',
        'name': 'TestOntology',
        'author': 'Author',
        'description': 'A test ontology',
        'type': 'Базовый',
        'date_created': '2024-01-01',
    }


def test_serialize_ontology_without_meta(serializer):
    ontology = Ontology()
    ontology.add_type(
        Term(name='MyType', description='A sample type', label='A some label')
    )
    ontology.add_function(
        Function(
            name='MyFunction',
            label='Function',
            input_types=['int'],
            output_type=['bool'],
        )
    )
    ontology.add_relationship(
        Relationship(
            relationship='',
            parent='',
            child='',
        )
    )

    json_output = serializer.serialize(ontology)
    data = json.loads(json_output)

    assert 'terms' in data
    assert 'functions' in data
    assert 'hierarchy' in data
    assert data['meta'] == {
        'version': None,
        'name': None,
        'author': None,
        'description': None,
        'type': None,
        'date_created': None,
    }
    assert data['terms'] == [
        {
            'name': 'MyType',
            'label': 'A some label',
            'description': 'A sample type',
            'attributes': {},
        }
    ]
    assert data['functions'] == [
        {
            'name': 'MyFunction',
            'label': 'Function',
            'input_types': ['int'],
            'output_type': ['bool'],
            'attributes': {},
        }
    ]
    assert data['hierarchy'] == [
        {'parent': '', 'relationship': '', 'child': '', 'attributes': {}}
    ]



from src.ontol import Parser, Term, Function, Relationship

import pytest


@pytest.fixture
def parser():
    return Parser()


def test_parse_empty_file(parser):
    content: str = ''
    ontology, warnings = parser.parse(content, 'test.ontol')
    assert ontology.meta.version is None
    assert ontology.meta.name is None
    assert ontology.meta.author is None
    assert ontology.meta.description is None
    assert ontology.meta.type is None
    assert ontology.meta.date_created is not None
    assert len(ontology.types) == 0
    assert len(ontology.functions) == 0
    assert len(ontology.hierarchy) == 0
    assert len(warnings) == 0


def test_parse_commented_file(parser):
    content: str = """
    # version: '1.0'
    # title: 'Basic linear algebra'
    # author: 'Firstname Lastname'
    # desc: 'Matrices'

    # types:
    # number: 'Number', ''
    # matrix: 'Matrix', ''

    # functions:
    # transpose: 'Transpose' (matrix) -> matrix: 'Transposed matrix'

    # hierarchy:
    # matrix composition number
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert ontology.meta.version is None
    assert ontology.meta.name is None
    assert ontology.meta.author is None
    assert ontology.meta.description is None
    assert len(ontology.types) == 0
    assert len(ontology.functions) == 0
    assert len(ontology.hierarchy) == 0
    assert len(warnings) == 0


def test_parse_type(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов', {color: '#ffffff'}
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == 'Множество'
    assert term.description == 'Коллекция уникальных элементов'
    assert term.attributes == {'color': '#ffffff'}


def test_parse_type_without_arguments(parser):
    content: str = """
    types:
    set: 'Множество', 'Коллекция уникальных элементов'
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == 'Множество'
    assert term.description == 'Коллекция уникальных элементов'
    assert term.attributes == {}
    assert len(warnings) == 0


def test_parse_type_with_empty_label_and_desc(parser):
    content: str = """
    types:
    set: '', ''
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 1

    term: Term = ontology.types[0]
    assert term.name == 'set'
    assert term.label == ''
    assert term.description == ''
    assert term.attributes == {}
    assert len(warnings) == 2


def test_parse_type_without_label_and_desc(parser):
    content: str = """
    types:
    set: ''
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')

    content: str = """
    types:
    set
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_parse_function(parser):
    content = """
    functions:
    descartes: 'Cartesian product' (set: 'First set', set: 'Second set') -> set: 'Result set', { color: '#fff }
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.functions) == 1

    func: Function = ontology.functions[0]
    assert func.name == 'descartes'
    assert func.label == 'Cartesian product'
    assert func.input_types == [('set', 'First set'), ('set', 'Second set')]
    assert func.output_type == ('set', 'Result set')
    assert func.attributes == {'color': '#fff'}
    assert len(warnings) == 0


def test_parse_function_with_empty_attributes(parser):
    content = """
    functions:
    divide: 'Divide two numbers' (number: 'Dividend', number: 'Divisor') -> number: 'Quotient', {}
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.functions) == 1

    func: Function = ontology.functions[0]
    assert func.name == 'divide'
    assert func.label == 'Divide two numbers'
    assert func.input_types == [('number', 'Dividend'), ('number', 'Divisor')]
    assert func.output_type == ('number', 'Quotient')
    assert func.attributes == {}
    assert len(warnings) == 0


def test_parse_function_with_missing_output(parser):
    content = """
    functions:
    add: 'Addition' (number: 'First number', number: 'Second number') -> , {color: '#fff'}
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_parse_function_with_missing_label(parser):
    content = """
    functions:
    add(number: 'First number', number: 'Second number') -> number: 'Result', {color: '#fff'}
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_parse_function_without_arguments(parser):
    content = """
    functions:
    today: 'Returns current date' () -> date: 'Current date'
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.functions) == 1

    func: Function = ontology.functions[0]
    assert func.name == 'today'
    assert func.label == 'Returns current date'
    assert func.input_types == []
    assert func.output_type == ('date', 'Current date')
    assert len(warnings) == 0


def test_parse_with_empty_labels(parser):
    content = """
    functions:
    descartes: '' (set: '', set: '') -> set: ''
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.functions) == 1

    func: Function = ontology.functions[0]
    assert func.name == 'descartes'
    assert func.label == ''
    assert func.input_types == [('set', ''), ('set', '')]
    assert func.output_type == ('set', '')
    assert len(warnings) == 4


def test_parse_heierarchy(parser):
    content = """
    hierarchy:
    element inheritance set
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.hierarchy) == 1

    rel: Relationship = ontology.hierarchy[0]
    assert rel.parent == 'element'
    assert rel.relationship == 'inheritance'
    assert rel.child == ['set']
    assert len(warnings) == 0


def test_parse_meta(parser):
    content = """
    version: '1.0'
    title: "Basic calculus" # can we come up with a more interesting name?
    author: 'Firstname Lastname'
    desc: "Limits, differentiation and integrals"
    type: 'Базовый'
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert ontology.meta is not None
    assert ontology.meta.version == '1.0'
    assert ontology.meta.name == 'Basic calculus'
    assert ontology.meta.author == 'Firstname Lastname'
    assert ontology.meta.description == 'Limits, differentiation and integrals'
    assert ontology.meta.type == 'Базовый'
    assert len(warnings) == 0


def test_parse_meta_without_qoutes(parser):
    content = """
    version: 1.0
    title: "Basic calculus"
    author: 'Firstname Lastname'
    desc: "Limits, differentiation and integrals"
    """
    with pytest.raises(SyntaxError):
        parser.parse(content, 'test.ontol')


def test_combined_parsing(parser):
    content = """
    # Ontology information
    version: '1.0'
    title: 'Basic linear algebra'

    # Types we have
    types:
    number: 'Number', 'Some real number from R field'

    # Other type
    types:
    matrix: 'Matrix', 'Rectangular array or table of numbers'

    # Functions we have
    functions:
    # Describe more functions
    transpose: 'Transpose' (matrix) -> matrix: 'Transposed matrix'

    # And relations we have
    hierarchy:
    matrix composition number
    """
    ontology, warnings = parser.parse(content, 'test.ontol')

    assert len(ontology.types) == 2
    assert len(ontology.functions) == 1
    assert len(ontology.hierarchy) == 1

    assert ontology.meta is not None

    assert len(warnings) == 0



import re

from typing import Optional, Literal
from ontol import Ontology, Term, Function, Relationship, Meta
from datetime import datetime


class Parser:
    def __init__(self):
        self.warnings: list[str] = []

    def parse(self, file_content: str, file_path: str) -> tuple[Ontology, list]:
        self.warnings.clear()

        ontology: Ontology = Ontology()

        lines: list[str] = file_content.splitlines()
        current_block: Optional[Literal['types', 'functions', 'hierarchy']] = None
        meta_data: dict[str, Optional[str]] = {
            'version': None,
            'name': None,
            'author': None,
            'description': None,
            'type': None,
            'date_created': datetime.today().strftime('%Y-%m-%d'),
        }

        for index, line in enumerate(lines):
            line_number: int = index + 1
            line = line.strip()

            if line.startswith('#') or not line:
                continue

            try:
                if re.match(r'^version:\s+', line):
                    meta_data['version'] = self._parse_meta_line(
                        line, 'version', file_path, line_number
                    )
                elif re.match(r'^title:\s+', line):
                    meta_data['name'] = self._parse_meta_line(
                        line, 'title', file_path, line_number
                    )
                elif re.match(r'^author:\s+', line):
                    meta_data['author'] = self._parse_meta_line(
                        line, 'author', file_path, line_number
                    )
                elif re.match(r'^desc:\s+', line):
                    meta_data['description'] = self._parse_meta_line(
                        line, 'desc', file_path, line_number
                    )
                elif re.match(r'^type:\s+', line):
                    meta_data['type'] = self._parse_meta_line(
                        line, 'type', file_path, line_number
                    )

                elif line.startswith('types:'):
                    current_block = 'types'
                    continue
                elif line.startswith('functions:'):
                    current_block = 'functions'
                    continue
                elif line.startswith('hierarchy:'):
                    current_block = 'hierarchy'
                    continue

                elif current_block == 'types':
                    type_def = self._parse_type(line, file_path, line_number)
                    ontology.add_type(type_def)
                elif current_block == 'functions':
                    func_def = self._parse_function(line, file_path, line_number)
                    ontology.add_function(func_def)
                elif current_block == 'hierarchy':
                    relationship = self._parse_relationship(line)
                    ontology.add_relationship(relationship)

                else:
                    raise SyntaxError('Unexpected line')

            except Exception as e:
                raise SyntaxError(
                    f'File "{file_path}", line {line_number}\n    {line}\n\033[31m{type(e).__name__}: \033[0m{e}'
                )

        ontology.set_meta(Meta(**meta_data))

        return ontology, self.warnings

    def _parse_meta_line(
        self, line: str, key: str, file_path: str, line_number: int
    ) -> str:
        match = re.match(rf"{key}:\s+['\"](.*?)['\"]", line)

        if not match:
            raise SyntaxError(f'Invalid {key} format')

        value = match.group(1)

        if not value:
            self._add_warning(file_path, line_number, line, 'Meta value is empty')

        return value

    def _parse_type(self, line: str, file_path: str, line_number: int) -> Term:
        match = re.match(
            r"(\w+):\s*['\"](.*?)['\"],\s*['\"](.*?)['\"](,\s*\{(.*?)\})?$", line
        )

        if not match:
            raise SyntaxError('Invalid type format')

        name: str = match.group(1)
        label: str = match.group(2)
        description: str = match.group(3)

        if not label:
            self._add_warning(file_path, line_number, line, 'Label is empty')
        if not description:
            self._add_warning(file_path, line_number, line, 'Description is empty')

        attributes: dict[str, str] = (
            self._parse_attributes(match.group(5)) if match.group(5) else {}
        )

        return Term(name, label, description, attributes)

    def _parse_attributes(self, attr_string: str) -> dict[str, str]:
        attributes: dict[str, str] = {}

        if attr_string:
            attr_string = attr_string.strip('{}').strip()
            for attr in attr_string.split(','):
                key, value = attr.split(':', 1)
                attributes[key.strip()] = value.strip().strip('\'"')

        return attributes

    def _parse_function(self, line: str, file_path: str, line_number: int) -> Function:
        match = re.match(
            r"(\w+):\s*['\"](.*?)['\"]\s*\((.*?)\)\s*->\s*(\w+):\s*['\"](.*?)['\"](,\s*\{(.*?)\})?$",
            line,
        )

        if not match:
            raise SyntaxError('Invalid function format')

        name: str = match.group(1)
        label: str = match.group(2)
        input_params: list[tuple[str, str]] = self._parse_parameters(
            match.group(3), line, file_path, line_number
        )
        output_type: str = match.group(4)
        output_label: str = match.group(5)
        attributes: dict[str, str] = (
            self._parse_attributes(match.group(7)) if match.group(7) else {}
        )

        if not label:
            self._add_warning(file_path, line_number, line, 'Label is empty')
        if not output_label:
            self._add_warning(file_path, line_number, line, 'Output label is empty')

        return Function(
            name, label, input_params, (output_type, output_label), attributes
        )

    def _parse_parameters(
        self, params_string: str, line: str, file_path: str, line_number: int
    ) -> list[tuple[str, str]]:
        params: list[tuple[str, str]] = []

        for param in params_string.split(','):
            param = param.strip()
            match = re.match(r"(\w+):\s*['\"](.*?)['\"]$", param)

            if match:
                param_name: str = match.group(1)
                param_label: str = match.group(2)

                if not param_label:
                    self._add_warning(
                        file_path,
                        line_number,
                        line,
                        f"{param_name}'s parameter label is empty",
                    )

                params.append((param_name, param_label))

        return params

    def _parse_relationship(self, line: str) -> Relationship:
        pattern = r'^(\w+)\s+(\w+)\s+(?:\(([^)]+)\)|(\w+))?\s*(?:\{([^}]*)\})?$'
        match = re.match(pattern, line.strip())
        if not match:
            raise ValueError(f'Invalid line {line}')

        parent: str = match.group(1)
        relation: str = match.group(2)
        child: str = match.group(3) or match.group(4)
        children: list[str] = child.split(', ') if child else []

        attributes = match.group(5)
        attributes_dict = {
            'color': '#black',
            'title': '',
            'leftChar': '',
            'rightChar': '',
            'direction': 'forward',
        }

        if attributes:
            for item in attributes.split(', '):
                key, value = item.split(': ', 1)
                attributes_dict[key.strip()] = value.strip().strip('"\'')

        return Relationship(parent, relation, children, attributes_dict)

    def _add_warning(
        self, file_path: str, line_number: int, line: str, message: str
    ) -> None:
        warning: str = f'File "{file_path}", line {line_number}\n    {line}\n\033[33mWarning: \033[0m{message}'
        self.warnings.append(warning)



from .oast import Ontology, Term, Function, Relationship, Meta
from .parser import Parser
from .plantuml import PlantUML
from .serializer import JSONSerializer
from .cli import CLI


__all__ = (
    'Ontology',
    'Term',
    'Function',
    'Relationship',
    'Meta',
    'Parser',
    'PlantUML',
    'JSONSerializer',
    'CLI',
)



from typing import Optional, Dict
from dataclasses import dataclass, field


class ASTNode:
    pass


@dataclass
class Term(ASTNode):
    name: str
    label: str
    description: str
    attributes: dict = field(default_factory=dict)

    def __repr__(self) -> str:
        return f'Term(name={self.name}, label={self.label}, description={self.description}, attributes={self.attributes})'


@dataclass
class Function(ASTNode):
    name: str
    label: str
    # TODO: replace tuple[str, str] with list[{ name: str, label: str }]
    # TODO: replace str with Term
    input_types: list[tuple[str, str]]
    # TODO: replace tuple[str, str] with dict { name: str, label: str }
    # TODO: replace str with Term
    output_type: tuple[str, str]
    attributes: dict = field(default_factory=dict)

    def __repr__(self) -> str:
        return (
            f'Function(name={self.name}, label={self.label}, input_types={self.input_types},'
            f' output_type={self.output_type}, '
            f'attributes={self.attributes})'
        )


# TODO: implement structure from technical task. Must contain parent, child, and relationship type
@dataclass
class Relationship(ASTNode):
    # TODO: replace str with Term
    parent: str
    # TODO: replace str with enum type
    relationship: str
    # TODO: replace str with Term
    child: list[str]
    attributes: Dict[str, str] = field(default_factory=dict)

    def __repr__(self) -> str:
        return (
            f'Relationship(parent={self.parent}, '
            f'relationship={self.relationship}, '
            f'child={self.child}, attributes={self.attributes})'
        )


@dataclass
class Meta(ASTNode):
    version: Optional[str] = None
    # TODO: rename name to title
    name: Optional[str] = None
    author: Optional[str] = None
    description: Optional[str] = None
    type: Optional[str] = None
    date_created: Optional[str] = None

    def __repr__(self) -> str:
        return (
            f'Meta(version={self.version}, name={self.name}, author={self.author}, '
            f'description={self.description}, type={self.type}, '
            f'date_created={self.date_created})'
        )


@dataclass
class Ontology(ASTNode):
    types: list[Term] = field(default_factory=list)
    functions: list[Function] = field(default_factory=list)
    hierarchy: list[Relationship] = field(default_factory=list)
    meta: Meta = field(default_factory=Meta)

    def add_type(self, type_def: Term) -> None:
        self.types.append(type_def)

    def add_function(self, func_def: Function) -> None:
        self.functions.append(func_def)

    def add_relationship(self, logical_expr: Relationship) -> None:
        self.hierarchy.append(logical_expr)

    def set_meta(self, meta: Meta) -> None:
        self.meta = meta

    def __repr__(self) -> str:
        return (
            f'Ontology(types={self.types}, '
            f'functions={self.functions},'
            f'hierarchy={self.hierarchy},'
            f'meta={self.meta})'
        )
